#!/usr/bin/env python

'''

 pdf-analyzer is a simple tool to analyze suspicious PDF files which can contain obfuscated JS code
 and shellcodes, trying to retrieve them.

 Author: Jose Miguel Esparza    <josemiguel.esparza at gmail.com>
                                <jesparza at s21sec.com>   
 
 Timeline:
    2010-01-10 Added some JS deobfuscation tricks
    2009-06-02 Publication
 
 Requirements:

    - Pyrex
    - Spidermonkey
    
'''

import sys,zlib,os,re
from spidermonkey import Runtime

def searchEvents(content):
    for event in suspiciousEvents:
        if content.find(event) != -1:
            if event not in eventsFound:
                eventsFound.append(event)
                
def printEvents():
    eventsString = 'Trigger events: '
    if eventsFound != []:
        for event in eventsFound:
            eventsString += '\n\t' + event
    else:
        eventsString += '\n\tNot found!!'
    print eventsString + '\n'
            
def searchActions(content):
    for action in suspiciousActions:
        if content.find(action) != -1:
            if action not in actionsFound:
                actionsFound.append(action)

def printActions():
    actionsString = 'Suspicious actions: '
    if actionsFound != []:
        for action in actionsFound:
            actionsString += '\n\t' + action
    else:
        actionsString += '\n\tNot found!!'
    print actionsString + '\n'
            
def searchVulnFunctions(content):
    for function in vulnJSFunctions:
        if content.find(function) != -1:
            if function not in vulnFunctionsFound:
                vulnFunctionsFound.append(function)

def searchVulns(content):
    for vuln in knownVulns:
        if content.find(vuln) != -1:
            if vuln not in vulnFunctionsFound:
                vulnFunctionsFound.append(vuln)
                
def printVulns():
    vulnsString = 'Vulnerable elements: '
    if vulnFunctionsFound != []:
        for vuln in vulnFunctionsFound:
            vulnsString += '\n\t' + vuln
    else:
        vulnsString += '\n\tNot found!!'
    print vulnsString + '\n'
    
def printURLs():
    if urlsFound != []:
        print 'URLs in shellcode:'
        for url in urlsFound:
            print '\t'+url
            
                
def printBytes():
    row = 16
    print 'Unescaped bytes: '
    if unescapedBytes != []:
        print
        for bytes in unescapedBytes:
            i = None
            hexChain = ''
            strings = ''
            for i in range(0,len(bytes)):
                if ord(bytes[i]) > 31 and ord(bytes[i]) < 128:
                    strings += bytes[i]
                else:
                    strings += '.'
                hexChars = hex(ord(bytes[i]))
                hexChars = hexChars[2:]
                if len(hexChars) == 1:
                    hexChars = '0' + hexChars
                hexChain += hexChars + ' '
                if i != 0 and i % row == row -1:
                    print hexChain + '  |' + strings + '|'
                    hexChain = ''
                    strings = ''
            if i != None and i % row != 0:
                if hexChain == '':
                    print '\n\n'
                else:
                    print hexChain + (48 - len(hexChain))*' ' + '  |' + strings + '|\n\n'
    else:
        print '\tNot found!!\n'
        
def writeFiles():
    countJS = 0
    countBin = 0
    name = os.path.basename(fileName)
    name = name[:-4]
    # JS code
    for js in JSCode:
        open(name + '-' + str(countJS) + '.js','w').write(js)
        countJS += 1
    # Raw bytes
    for bytes in unescapedBytes:
        open(name + '-' + str(countBin) + '.bin','wb').write(bytes)
        countBin += 1

def cleanStream(stream, length):
   output = stream
   garbage = len(stream) - length
   if garbage == 0 or garbage < 0:
       return stream
   elif garbage > 0:  
       for i in range(len(stream)):
           if garbage == 0:
               break
           if stream[i] == '\r' or stream[i] == '\n':
               output = output[1:]
               garbage -= 1
           else:
               break
       for i in range(len(stream)-1,0,-1):
           if garbage == 0:
               break
           if stream[i] == '\r' or stream[i] == '\n':
               output = output[:-1]
               garbage -= 1
           else:
               break
       return output

def isPDF(content):
    firstLine = content[:content.find('\n')]
    if re.findall('%PDF-\d\.\d' ,firstLine, re.DOTALL):
        return True
    else:
        return False

def isJavascript(content):
    JSStrings = ['var ',';',')','(','function ','=','{','}','if ','else','return','while ','for ',',','eval']
    keyStrings = [';','(',')']
    stringsFound = []
    limit = 15
    results = 0
    
    for char in content:
        if ord(char) >= 128:
            return False

    for string in JSStrings:
        cont = content.count(string)
        results += cont
        if cont > 0:
            stringsFound.append(string)
        elif cont == 0 and string in keyStrings:
			return False

    if results > limit and len(stringsFound) >= 5:
        return True
    else:
        return False
    
def unescape(escapedBytes):
    global urlsFound 
    unescapedBytes = ''
    for i in range(2,len(escapedBytes)-1,6):
		unescapedBytes += chr(int(escapedBytes[i+2]+escapedBytes[i+3],16))+chr(int(escapedBytes[i]+escapedBytes[i+1],16))
    urls = re.findall('https?://.*$', unescapedBytes, re.DOTALL)
    for url in urls:
       if url not in urlsFound:
          urlsFound.append(url)
    return unescapedBytes

def getVarContent(js, bytes):
    clearBytes = ''
    bytes = bytes.replace('\n','')
    bytes = bytes.replace('\r','')
    bytes = bytes.replace('\t','')
    bytes = bytes.replace(' ','')
    parts = bytes.split('+')
    for part in parts:
        if re.match('["\'].*?["\']', part, re.DOTALL):
            clearBytes += part[1:-1]
        else:
            varContent = re.findall(part + '\s*?=\s*?(.*?)[,;]', js, re.DOTALL)
            if varContent != []:
                clearBytes += getVarContent(js, varContent[0])
    return clearBytes


def searchObfuscatedFunctions(jsCode, function):
    obfuscatedFunctionsInfo = []
    match = re.findall('\W('+function+'\s{0,5}?\((.*?)\)\s{0,5}?;)', jsCode, re.DOTALL)
    if match != []:
       for m in match:
          if re.findall('return',m[1],re.IGNORECASE) != []:
             obfuscatedFunctionsInfo.append([function,m,True])
          else:
             obfuscatedFunctionsInfo.append([function,m,False])
    obfuscatedFunctions = re.findall('\s*?((\w*?)\s*?=\s*?'+function+')\s*?;', element, re.DOTALL)
    for obfuscatedFunction in obfuscatedFunctions:
       obfuscatedElement = obfuscatedFunction[1]
       obfuscatedFunctionsInfo += searchObfuscatedFunctions(jsCode, obfuscatedElement)
    return obfuscatedFunctionsInfo


suspiciousEvents = ['/OpenAction','/AA']
eventsFound = []
suspiciousActions = ['/JS','/Launch','/SubmitForm','/ImportData']
actionsFound = []
vulnJSFunctions = ['mailto','Collab.collectEmailInfo','util.printf','getAnnots','getIcon','spell.customDictionaryOpen']
knownVulns = ['/JBIG2Decode']
vulnFunctionsFound = []
JSCode = []
elementsToAnalyze = []
unescapedBytes = []
urlsFound = []
writeMode = False
usage = 'Usage: '+sys.argv[0]+' [-w] file\n\nOptions:\n\t-w: write JS code and shellcode to disk'


if len(sys.argv) != 2 and len(sys.argv) != 3:
    sys.exit(usage)

if len(sys.argv) == 2:
    fileName = sys.argv[1]
else:
    param = sys.argv[1]
    fileName = sys.argv[2]
    if param == '-w':
        writeMode = True
    else:
        sys.exit(usage)
        
if os.path.exists(fileName):
   content = open(fileName,'r').read()
   if not isPDF(content):
     print 'Error: PDF header not found!!'
     res = raw_input('Do you want to continue? (y/n) ')
     if res.lower() != 'y':
       sys.exit()
   searchEvents(content)
   searchActions(content)
   searchVulnFunctions(content)
   searchVulns(content)
   
   if '/JS' in actionsFound:
       jsContent = re.findall('/JS\s*?\((.*?)\)[\r\n]', content, re.DOTALL)
       if jsContent != []:
           js = jsContent[0]
           js = js.replace('\(','(')
           js = js.replace('\)',')')
           js = js.replace('\n',' ')
           js = js.replace('\r',' ')
           js = js.replace('\t',' ')
           js = js.replace('\\\\','\\')
           elementsToAnalyze.append(js)
                    
   streams = re.findall('obj\s{0,10}<<(.{0,100}?/Length\s([0-9]{1,8}).*?)>>[\s%]*?stream(.*?)endstream',content,re.DOTALL)   
   for streamDict in streams:
       dict = streamDict[0]
       if dict.count('obj') > 0:
           dict = dict[dict.rfind('obj'):]
       stream = cleanStream(streamDict[2],int(streamDict[1]))
       if dict.find('/FlateDecode') != -1 or dict.find('/Fl') != -1:
           try:
               streamContent = zlib.decompress(stream)
           except:
               continue
       else:
           streamContent = stream
       elementsToAnalyze.append(streamContent)

   for element in elementsToAnalyze:
        if isJavascript(element):
            JSCode.append(element)
            r = Runtime()
            context = r.new_context()
            while True:
                 evalFunctionsData = searchObfuscatedFunctions(element, 'eval')
                 originalElement = element
                 for evalFunctionData in evalFunctionsData:
                     if not evalFunctionData[2]:
                        modifiedCode = evalFunctionData[1][0].replace(evalFunctionData[0],'return')
                        element = originalElement.replace(evalFunctionData[1][0],modifiedCode)
                     else:
                        element = originalElement.replace(evalFunctionData[1][0],evalFunctionData[1][1]+';')
                     try:
                        executedJS = context.eval_script(element)
                        if executedJS == None:
                           raise exception
                        break
                     except:
                        if evalFunctionData[2]:
                           modifiedCode = evalFunctionData[1][0].replace(evalFunctionData[0],'return')
                           element = originalElement.replace(evalFunctionData[1][0],modifiedCode)
                        else:
                           element = originalElement.replace(evalFunctionData[1][0],evalFunctionData[1][1]+';')
                        try:
                           executedJS = context.eval_script(element)
                           break
                        except:
                           element = originalElement
                           continue
                 else:
                     break
                 if executedJS != originalElement:
                     element = executedJS
                     JSCode.append(element)                
                 else:                                            
                     break
            searchVulnFunctions(element)
            escapedVars = re.findall('(\w*?)\s*?=\s*?(unescape\((.*?)\))', element, re.DOTALL)
            for var in escapedVars:
                bytes = var[2]
                if bytes.find('+') != -1:
                    varContent = getVarContent(element, bytes)
                    if len(varContent) > 150:
                        bytes = unescape(varContent)
                        if bytes not in unescapedBytes:
                           unescapedBytes.append(bytes)
                else:
                    bytes = bytes[1:-1]
                    if len(bytes) > 150:
                        bytes = unescape(bytes)
                        if bytes not in unescapedBytes:
                           unescapedBytes.append(bytes)
                 
   print 'File: ' + fileName + '\n'
   printEvents()
   printActions()
   printVulns()
   printBytes()
   printURLs()
   if writeMode:
       writeFiles()
else:
    sys.exit('Error: the file doesn\'t exist!!')
